"""Sphinx configuration."""

project = "Inductance"
author = "Darren Garnier"
copyright = "2023, Darren Garnier"
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "myst_parser",
]
autodoc_typehints = "description"
html_theme = "furo"



"""Test the mutual inductance of filaments and paths."""

import unittest

import coverage_env  # noqa: F401
from numpy import array, cos, dstack, linspace, ones_like, pi, sin

from inductance.filaments import (
    M_path_path,
    _loop_segmented_mutual,
    mutual_inductance_fil,
)


class TestMutual(unittest.TestCase):
    """Test and compare mutual inductance of filaments and paths."""

    def setUp(self) -> None:
        """Set up the test data."""
        self.coplanar_fil1 = array([1, 10, 1])
        self.coplanar_fil2 = array([100, 10, 1])
        t = linspace(0, 2 * pi, 10000)
        self.coplanar_fil1_path = dstack((-1 * sin(t), 1 * cos(t), 10 * ones_like(t)))[
            0
        ]
        self.coplanar_fil2_path = dstack(
            (-100 * sin(t), 100 * cos(t), 10 * ones_like(t))
        )[0]

        self.coaxial_fil1 = array([1, 0, 1])
        self.coaxial_fil2 = array([1, 100, 1])
        self.coaxial_fil1_path = dstack((-sin(t), cos(t), 0 * ones_like(t)))[0]
        self.coaxial_fil2_path = dstack((-sin(t), cos(t), 100 * ones_like(t)))[0]

        mu_0 = 1.256637062e-6  # H/m
        self.coplanar_analytic_sol = (
            mu_0 * pi * self.coplanar_fil1[0] ** 2 / (2 * self.coplanar_fil2[0])
        )
        self.coaxial_analytic_sol = (
            mu_0
            * pi
            * self.coaxial_fil1[0] ** 4
            / (2 * abs(self.coaxial_fil2[1] - self.coaxial_fil1[1]) ** 3)
        )
        return super().setUp()

    def test_mutual_fil_fil(self):
        """Test filaments.mutual_inductance_fil."""
        coplanar_filaments_sol = mutual_inductance_fil(
            self.coplanar_fil1, self.coplanar_fil2
        )
        coaxial_filaments_sol = mutual_inductance_fil(
            self.coaxial_fil1, self.coaxial_fil2
        )

        self.assertAlmostEqual(
            self.coplanar_analytic_sol, coplanar_filaments_sol, places=11
        )
        self.assertAlmostEqual(
            self.coaxial_analytic_sol, coaxial_filaments_sol, places=11
        )

    def test_mutual_fil_path(self):
        """Test filaments._loop_segmented_mutual."""
        coplanar_path_sol = _loop_segmented_mutual(
            *self.coplanar_fil1[:2], self.coplanar_fil2_path
        )
        coaxial_path_sol = _loop_segmented_mutual(
            *self.coaxial_fil1[:2], self.coaxial_fil2_path
        )

        self.assertAlmostEqual(coplanar_path_sol, self.coplanar_analytic_sol, places=11)
        self.assertAlmostEqual(coaxial_path_sol, self.coaxial_analytic_sol, places=11)

    def test_mutual_path_path(self):
        """Test filaments.M_path_path."""
        coplanar_path_sol = M_path_path(
            self.coplanar_fil1_path, self.coplanar_fil2_path
        )
        coaxial_path_sol = M_path_path(self.coaxial_fil1_path, self.coaxial_fil2_path)

        self.assertAlmostEqual(coplanar_path_sol, self.coplanar_analytic_sol, places=10)
        self.assertAlmostEqual(coaxial_path_sol, self.coaxial_analytic_sol, places=10)


if __name__ == "__main__":
    unittest.main()



"""Coil inductance calculations.

Defines a coil class to keep track of coil parameters.

Benchmarking against LDX values, which come from
old Mathematica routines and other tests.

Filaments are defined as an numpy 3 element vector
 - r, z, and n.
"""

from __future__ import annotations

import math
from dataclasses import dataclass
from enum import Enum

import numpy as np

from .filaments import (
    filament_coil,
    mutual_inductance_of_filaments,
    radial_force_of_filaments,
    vertical_force_of_filaments,
)
from .self import (
    L_long_solenoid_butterworth,
    L_lorentz,
    L_lyle4,
    L_lyle6,
    L_lyle6_appendix,
    L_maxwell,
    dLdR_lyle6,
    self_inductance_by_filaments,
)


class Shape(Enum):
    """Enum for conductor shapes."""

    round = "round"
    hollow_round = "hollow_round"
    rect = "rectangle"


@dataclass
class Conductor:
    """Conductor object to keep track of conductor parameters."""

    shape: Shape | str = Shape.round
    r: float = 0.0
    dr: float = 0.0
    dz: float = 0.0
    r_i: float = 0.0
    rho: float = 0.0

    def __post_init__(self):
        """Post init."""
        if isinstance(self.shape, str):
            self.shape = Shape(self.shape)


class Coil:
    """Rectangular coil object to keep track of coil parameters."""

    def __init__(self, r, z, dr, dz, nt=1, at=None, nr=0, nz=0, theta=0, **kwargs):
        """Create a rectangular coil object.

        Args:
            r (float): radial center of coil
            z (float): vertical center of coil
            dr (float): radial width of coil
            dz (float): axial height of coil
            nt (int, optional): number of turns in coil
            nr (int, optional): Number of radial sections to filament coil. Defaults to 0.
            nz (int, optional): Number of axial sections to filament coil. Defaults to 0.
            at (float, optional): Amperage of coil. Defaults to nt Amps.
            theta (float, optional): Rotation angle in radians. Defaults to 0.
            **kwargs: Additional arguments to store in the coil object.
        """
        self.r = r
        self.z = z
        self.dr = dr
        self.dz = dz
        self.nt = nt
        if at is None:
            at = nt
        self.at = at
        self.theta = theta
        self.fils = None
        self._L = None
        self._r_c = None
        self._z_c = None

        if kwargs:
            for key, value in kwargs.items():
                setattr(self, key, value)

        if (nr > 0) and (nz > 0):
            self.nr = nr
            self.nz = nz
            self.filamentize(nr, nz)

    @classmethod
    def from_dict(cls, d):
        """Create a coil from a dictionary."""
        if "r1" in d:
            return cls.from_bounds(**d)
        else:
            return cls(**d)

    @classmethod
    def from_bounds(cls, r1, r2, z1, z2, nt=1, at=1, nr=0, nz=0):
        """Create a coil from bounds instead of center and width."""
        return cls(
            (r1 + r2) / 2, (z1 + z2) / 2, r2 - r1, z2 - z1, nt=nt, at=at, nr=nr, nz=nz
        )

    @property
    def r1(self):  # noqa: D102
        return self.r - self.dr / 2

    @property
    def r2(self):  # noqa: D102
        return self.r + self.dr / 2

    @property
    def z1(self):  # noqa: D102
        return self.z - self.dz / 2

    @property
    def z2(self):  # noqa: D102
        return self.z + self.dz / 2

    @property
    def r_c(self):  # noqa: D102
        return self.r if self._r_c is None else self._r_c

    @property
    def z_c(self):  # noqa: D102
        return self.z if self._z_c is None else self._z_c

    @property
    def conductor_length(self):  # noqa: D102
        return self.r * self.nt * 2 * math.pi

    def L_best(self):
        """Inductance by best formula."""
        if self.dz < 0.2 * self.r:
            # Lyle's formula is not valid for long coils
            # as its an expansion in dz/r
            return self.L_Lyle6()
        else:
            return self.L_filament()

    @property
    def L(self):  # noqa: D102
        if self._L is None:
            self._L = self.L_best()
        return self._L

    @L.setter
    def L(self, value):
        self._L = value

    def filamentize(self, nr, nz):
        """Create an array of filaments to represent the coil."""
        self.nr = nr
        self.nz = nz
        self.fils = filament_coil(
            self.r, self.z, self.dr, self.dz, self.nt, nr, nz, theta=self.theta
        )

    def L_Maxwell(self):
        """Inductance by Maxwell's formula."""
        return L_maxwell(self.r, self.dr, self.dz, self.nt)

    def L_Lyle4(self):
        """Inductance by Lyle's formula, 4th order."""
        return L_lyle4(self.r, self.dr, self.dz, self.nt)

    def L_Lyle6(self):
        """Inductance by Lyle's formula, 6th order."""
        return L_lyle6(self.r, self.dr, self.dz, self.nt)

    def L_Lyle6A(self):
        """Inductance by Lyle's formula, 6th order, appendix."""
        return L_lyle6_appendix(self.r, self.dr, self.dz, self.nt)

    def L_filament(self, nr=0, nz=0):
        """Inductance by filamentation."""
        if nr != 0 and nz != 0:
            self.filamentize(nr, nz)
        cond = getattr(self, "conductor", None)
        if cond is None:
            return self_inductance_by_filaments(
                self.fils,
                conductor=Shape.rect.name,
                dr=self.dr / self.nr,
                dz=self.dz / self.nz,
            )
        elif "round" in cond.shape.value:
            return self_inductance_by_filaments(
                self.fils, conductor=cond.shape.name, a=cond.r
            )
        elif "rect" in cond.shape.value:
            return self_inductance_by_filaments(
                self.fils, conductor=cond.shape.name, dr=cond.dr, dz=cond.dz
            )

    def L_long_solenoid_butterworth(self):
        """Inductance by Butterworth's formula."""
        return L_long_solenoid_butterworth(self.r, self.dr, self.dz, self.nt)

    def L_lorentz(self):
        """Inductance by Lorentz's formula."""
        return L_lorentz(self.r, self.dr, self.dz, self.nt)

    def dLdR_Lyle6(self):
        """Derivative of inductance by Lyle's formula, 6th order."""
        return dLdR_lyle6(self.r, self.dr, self.dz, self.nt)

    def M_filament(self, C2: Coil) -> float:
        """Mutual inductance of two coils by filamentation."""
        return mutual_inductance_of_filaments(self.fils, C2.fils)

    def Fz_filament(self, C2: Coil) -> float:
        """Vertical force of two coils by filamentation."""
        F_a2 = vertical_force_of_filaments(self.fils, C2.fils)
        return self.at / self.nt * C2.at / C2.nt * F_a2

    def Fr_self(self) -> float:
        """Radial force of coil on itself."""
        dLdR = dLdR_lyle6(self.r, self.dr, self.dz, self.nt)
        return (self.at / self.nt) ** 2 / 2 * dLdR

    def Fr_filament(self, C2: Coil) -> float:
        """Radial force of two coils by filamentation."""
        F_r2 = radial_force_of_filaments(self.fils, C2.fils)
        return self.at / self.nt * C2.at / C2.nt * F_r2


class CompositeCoil(Coil):
    """A coil made of multiple rectangular coils."""

    def __init__(self, coils: list[Coil], **kwargs):
        """Create a composite coil from a list of _filamented_ coils."""
        self.coils = coils
        nt = sum(coil.nt for coil in coils)
        at = sum(coil.at for coil in coils)
        r1 = min(coil.r - coil.dr for coil in coils)
        r2 = max(coil.r + coil.dr for coil in coils)
        r = (r1 + r2) / 2
        dr = r2 - r1
        z1 = min(coil.z - coil.dz for coil in coils)
        z2 = max(coil.z + coil.dz for coil in coils)
        z = (z1 + z2) / 2
        dz = z2 - z1
        Coil.__init__(self, r, z, dr, dz, nt, at, **kwargs)
        self._r_c = sum(coil.r * coil.nt for coil in coils) / nt
        self._z_c = sum(coil.z * coil.nt for coil in coils) / nt
        self.fils = np.concatenate([coil.fils for coil in coils])

    @property
    def conductor_length(self):  # noqa: D102
        return sum(coil.conductor_length for coil in self.coils)

    def L_best(self):
        """Inductance of composite coils by best formula."""
        L = 0
        for i, c1 in enumerate(self.coils):
            for j, c2 in enumerate(self.coils):
                if i == j:
                    L += c1.L
                else:
                    L += c1.M_filament(c2)
        return L


def coilset_mutual_inductance(coils: list[Coil]):
    """Get the inductance matrix of a set of coils.

    Args:
        coils (list[Coil]): set of coils

    Returns:
        np.ndarray: 2D array of inductances
    """
    muts = np.zeros([len(coils), len(coils)], dtype=float)
    for i, ci in enumerate(coils):
        for j, cj in enumerate(coils):
            if i == j:
                muts[i, j] = ci.L
            else:
                muts[i, j] = ci.M_filament(cj)
    return muts


def coilset_Fz_greens(coils: list[Coil]):
    """Calculate the vertical force matrix of a set of coils.

    Args:
        coils (list[Coil]): set of coils

    Returns:
        np.ndarray: 2D array of vertical forces per amp**2
    """
    mfz = np.zeros([len(coils), len(coils)], dtype=float)
    for i, ci in enumerate(coils):
        for j, cj in enumerate(coils):
            if i != j:
                mfz[i, j] = ci.Fz_filament(cj)
    return mfz


def coilset_Fr_greens(coils: list[Coil]):
    """Calculate the radial force matrix of a set of coils.

    Args:
        coils (list[Coil]): set of coils

    Returns:
        np.ndarray: 2D array of radial forces per amp**2
    """
    muts = np.zeros([len(coils), len(coils)], dtype=float)
    for i, ci in enumerate(coils):
        for j, cj in enumerate(coils):
            if i == j:
                muts[i, j] = ci.Fr_self()
            else:
                muts[i, j] = ci.Fr_filament(cj)
    return muts



"""Mutual inductance calculations for coils.

author: Darren Garnier <garnier@mit.edu>


"""

import math

import numpy as np

from ._numba import njit
from .filaments import mutual_inductance_fil, mutual_inductance_of_filaments
from .utils import section_coil


@njit
def mutual_rayleigh(r1, z1, dr1, dz1, n1, r2, z2, dr2, dz2, n2):
    """Mutual inductance of two coils by Rayleigh's Quadrature Method.

    reproduced in :
    Rosa and Grover, "Formulas for the Mutual Inductance of
    Coaxial Circular Coils of Rectangular Section,"
    Bull. Natl. Bur. Stand., vol 8, no. 1, p. 34-35; 1911.

    Args:
        r1 (float): inner radius of coil 1
        z1 (float): inner height of coil 1
        dr1 (float): radial width of coil 1
        dz1 (float): height of coil 1
        n1 (int): number of turns in coil 1
        r2 (float): inner radius of coil 2
        z2 (float): inner height of coil 2
        dr2 (float): radial width of coil 2
        dz2 (float): height of coil 2
        n2 (int): number of turns in coil 2

    Returns:
        float: mutual inductance of the two coils
    """
    m_ray = 0
    # define the quadrature points
    rzn1 = np.array(
        [
            [r1, z1, 1],
            [r1 - dr1 / 2, z1, 1],
            [r1, z1 - dz1 / 2, 1],
            [r1 + dr1 / 2, z1, 1],
            [r1, z1 + dz1 / 2, 1],
        ]
    )
    rzn2 = np.array(
        [
            [r2, z2, 1],
            [r2 - dr2 / 2, z2, 1],
            [r2, z2 - dz2 / 2, 1],
            [r2 + dr2 / 2, z2, 1],
            [r2, z2 + dz2 / 2, 1],
        ]
    )
    # apply Rayleigh's Quadrature Method
    m_ray = -mutual_inductance_fil(rzn1[0, :], rzn2[0, :]) * 2
    m_ray += mutual_inductance_fil(rzn1[1, :], rzn2[0, :])
    m_ray += mutual_inductance_fil(rzn1[2, :], rzn2[0, :])
    m_ray += mutual_inductance_fil(rzn1[3, :], rzn2[0, :])
    m_ray += mutual_inductance_fil(rzn1[4, :], rzn2[0, :])
    m_ray += mutual_inductance_fil(rzn1[0, :], rzn2[1, :])
    m_ray += mutual_inductance_fil(rzn1[0, :], rzn2[2, :])
    m_ray += mutual_inductance_fil(rzn1[0, :], rzn2[3, :])
    m_ray += mutual_inductance_fil(rzn1[0, :], rzn2[4, :])
    return n1 * n2 * m_ray / 6


@njit
def lyle_equivalent_filaments(r, z, dr, dz, nt, fils):
    """Compute the equivalent filament locations for Lyle's method.

    Using Lyle's Method of Equivalent Filaments.

    originally from:
        Lyle, Phil. Mag., 3, p. 310; 1902.

    reproduced in:
        Rosa and Grover, "Formulas for the Mutual Inductance of
        Coaxial Circular Coils of Rectangular Section,"
        Bull. Natl. Bur. Stand., vol 8, no. 1, p. 38-39; 1911.

    Args:
        r (float): inner radius of coil
        z (float): inner height of coil
        dr (float): radial width of coil
        dz (float): height of coil
        nt (float): number of turns in coil
        fils (numpy.ndarray): array of filaments 2 x (r, z, n)
    """
    if dr < dz:
        req = r * (1 + dr**2 / (24 * r**2))
        beta = math.sqrt((dz**2 - dr**2) / 12)
        fils[:, 0] = req
        fils[:, 1] = z - beta, z + beta
    elif dr > dz:
        req = r * (1 + dz**2 / (24 * r**2))
        delta = math.sqrt((dr**2 - dz**2) / 12)
        fils[:, 0] = req - delta, req + delta
        fils[:, 1] = z
    else:
        req = r * (1 + dz**2 / (24 * r**2))
        fils[:, 0] = req
        fils[:, 1] = z
    fils[:, 2] = 0.5 * nt


@njit
def mutual_lyles_method(r1, z1, dr1, dz1, nt1, r2, z2, dr2, dz2, nt2):
    """Mutual inductance of two coils by Lyle's method.

    Using Lyle's Method of Equivalent Filaments.

    originally from:
    Lyle, Phil. Mag., 3, p. 310; 1902.

    reproduced:
    Rosa and Grover, "Formulas for the Mutual Inductance of
    Coaxial Circular Coils of Rectangular Section,"
    Bull. Natl. Bur. Stand., vol 8, no. 1, p. 38-39; 1911.

    Args:
        r1 (float): inner radius of coil 1
        z1 (float): inner height of coil 1
        dr1 (float): radial width of coil 1
        dz1 (float): height of coil 1
        nt1 (int): number of turns in coil 1
        r2 (float): inner radius of coil 2
        z2 (float): inner height of coil 2
        dr2 (float): radial width of coil 2
        dz2 (float): height of coil 2
        nt2 (int): number of turns in coil 2

    Returns:
        float: mutual inductance of the two coils
    """
    fils1 = np.zeros(2, 3)
    fils2 = np.zeros(2, 3)

    lyle_equivalent_filaments(r1, z1, dr1, dz1, nt1, fils1)
    lyle_equivalent_filaments(r2, z2, dr2, dz2, nt2, fils1)
    return mutual_inductance_of_filaments(fils1, fils2)


@njit
def lyle_equivalent_subcoil_filaments(subcoils):
    """Compute the equivalent filament locations for set of subcoils."""
    fils = np.zeros((subcoils.shape[0], 2, 3))
    for i in range(subcoils.shape[0]):
        lyle_equivalent_filaments(*subcoils[i, :], fils[i, :, :])
    return fils


def mutual_sectioning_lyle(
    r1, z1, dr1, dz1, nt1, nr1, nz1, r2, z2, dr2, dz2, nt2, nr2, nz2
):
    """Mutual inductance by sectioning of two coils by Lyle's Method.

    Section cois into subcoils and compute mutual inductance of each set
    of subcoil using Lyle's method of equivalent filaments.

    Args:
        r1 (float): inner radius of coil 1
        z1 (float): inner height of coil 1
        dr1 (float): radial width of coil 1
        dz1 (float): height of coil 1
        nt1 (float): number of turns in coil 1
        nr1 (int): number of radial sections in coil 1
        nz1 (int): number of vertical sections in coil 1
        r2 (float): inner radius of coil 2
        z2 (float): inner height of coil 2
        dr2 (float): radial width of coil 2
        dz2 (float): height of coil 2
        nt2 (float): number of turns in coil 2
        nr2 (int): number of radial sections in coil 2
        nz2 (int): number of vertical sections in coil 2

    Returns:
        float: mutual inductance of the two coils
    """
    subs_1 = section_coil(r1, z1, dr1, dz1, nt1, nr1, nz1)
    subs_2 = section_coil(r2, z2, dr2, dz2, nt2, nr2, nz2)

    fils_1 = lyle_equivalent_subcoil_filaments(subs_1).reshape(-1, 3)
    fils_2 = lyle_equivalent_subcoil_filaments(subs_2).reshape(-1, 3)

    return mutual_inductance_of_filaments(fils_1, fils_2)



"""Filamentary inductance calculations."""

import math

import numpy as np

from ._numba import guvectorize, njit, prange
from .elliptics import ellipke
from .utils import rectangle_GMD, section_coil


@njit
def mutual_inductance_fil(rzn1, rzn2):
    """Mutual inductance of two filaments.

    Args:
        rzn1 (array): (r, z, n) of first filament
        rzn2 (array): (r, z, n) of second filament

    Returns:
        float: mutual inductance in Henrys
    """
    r1, z1, n1 = rzn1
    r2, z2, n2 = rzn2

    k2 = 4 * r1 * r2 / ((r1 + r2) ** 2 + (z1 - z2) ** 2)
    elk, ele = ellipke(k2)
    amp = 2 * math.pi * r1 * 4e-7 * r2 / math.sqrt((r1 + r2) ** 2 + (z1 - z2) ** 2)
    M0 = n1 * n2 * amp * ((2 - k2) * elk - 2 * ele) / k2
    return M0


@njit
def vertical_force_fil(rzn1, rzn2):
    """Vertical force  between two filaments per conductor amp.

    Args:
        rzn1 (array): (r, z, n) of first filament
        rzn2 (array): (r, z, n) of second filament

    Returns:
        float: force in Newtons/Amp^2
    """
    r1, z1, n1 = rzn1
    r2, z2, n2 = rzn2

    BrAt1 = BrGreen(r1, z1, r2, z2)
    F = n1 * n2 * 2 * math.pi * r1 * BrAt1
    return F


@njit
def radial_force_fil(rzn1, rzn2):
    """Radial force  between two filaments per conductor amp.

    Args:
        rzn1 (array): (r, z, n) of first filament
        rzn2 (array): (r, z, n) of second filament

    Returns:
        float: force in Newtons/Amp^2
    """
    r1, z1, n1 = rzn1
    r2, z2, n2 = rzn2

    BzAt1 = BzGreen(r1, z1, r2, z2)
    F = n1 * n2 * 2 * math.pi * r1 * BzAt1
    return F


#
#  Green's functions for filaments
#


@njit
def AGreen(r, z, a, b):
    """Psi at position r, z, due to a filament with radius a, and z postion b.

    Args:
        r (float): radial position of a point
        z (float): vertical position of a point
        a (float): radius of a filament
        b (float): vertical position of a filament

    Returns:
        float: Psi at r, z in Weber / Amp
    """
    k2 = 4 * a * r / ((r + a) ** 2 + (z - b) ** 2)
    elk, ele = ellipke(k2)
    amp = 4e-7 * a / math.sqrt((r + a) ** 2 + (z - b) ** 2)
    return amp * ((2 - k2) * elk - 2 * ele) / k2


@njit
def ASegment(pts, xyz, uvw):
    """Psi at positions pts, due to a linear current segment uvw located at xyz.

    Derived from https://doi.org/10.1016/j.cpc.2023.108692, Eq. 20

    Args:
        pts (array): coordinates to calculate vector potential at, N x (x,y,z)
        xyz (array): coordinates of base of current element (x,y,z)
        uvw (array): displacement vector for length of element

    Returns:
        array: Psi at pts in Weber / Amp, N x (Psix,Psiy,Psiz)
    """
    mu02pi = 2e-7  # mu_0/2pi in H/m
    L = math.sqrt(uvw[0] ** 2 + uvw[1] ** 2 + uvw[2] ** 2)
    _uvw = uvw / L
    diffs = pts[:] - xyz
    zs = uvw[0] * diffs[:, 0] + uvw[1] * diffs[:, 1] + uvw[2] * diffs[:, 2]
    zvecs = zs[:, None] * _uvw
    rhovecs = diffs - zvecs
    rhos = np.sqrt(rhovecs[:, 0] ** 2 + rhovecs[:, 1] ** 2 + rhovecs[:, 2] ** 2) / L
    zs /= L

    ris = np.sqrt(rhos**2 + zs**2)
    rfs = np.sqrt(rhos**2 + (1 - zs) ** 2)
    epsilons = 1 / (ris + rfs)

    return mu02pi * np.arctanh(epsilons)[:, None] * _uvw


@njit
def BrGreen(r, z, a, b):
    """Br at position r, z, due to a filament with radius a, and z postion b.

    Args:
        r (float): radial position of a point
        z (float): vertical position of a point
        a (float): radius of a filament
        b (float): vertical position of a filament

    Returns:
        float: Br at r, z in Tesla / Amp
    """
    k2 = 4 * a * r / ((r + a) ** 2 + (z - b) ** 2)
    elk, ele = ellipke(k2)
    amp = -2e-7 / math.sqrt((r + a) ** 2 + (z - b) ** 2)
    Gr = (a**2 + r**2 + (z - b) ** 2) / ((r - a) ** 2 + (z - b) ** 2)
    return amp * (z - b) / r * (Gr * ele - elk)


@njit
def BzGreen(r, z, a, b):
    """Bz at position r, z, due to a filament with radius a, and z postion b.

    Args:
        r (float): radial position of a point
        z (float): vertical position of a point
        a (float): radius of a filament
        b (float): vertical position of a filament

    Returns:
        float: Bz at r, z in Tesla / Amp
    """
    k2 = 4 * a * r / ((r + a) ** 2 + (z - b) ** 2)
    elk, ele = ellipke(k2)
    amp = -2e-7 / math.sqrt((r + a) ** 2 + (z - b) ** 2)
    Gz = (a**2 - r**2 - (z - b) ** 2) / ((r - a) ** 2 + (z - b) ** 2)
    return amp * (Gz * ele + elk)


@njit(parallel=True)
def green_sum_over_filaments(gfunc, fil, r, z):
    """Calculate a greens function at position r, z, to an array of filaments.

    Args:
        gfunc (function): Green's function to use
        fil (array): filament array N x (r, z, n)
        r (float): radial position of a point
        z (float): vertical position of a point

    Returns:
        float: sum of the greens function at r, z, due to all filaments
    """
    gr = np.zeros_like(r)  # numba must compile this in object mode
    gf = gr.flat
    rf = r.flat
    zf = z.flat

    for j in prange(len(gr)):  # numba will njit this loop
        for i in range(fil.shape[0]):
            gf[j] += fil[i, 2] * gfunc(rf[j], zf[j], fil[i, 0], fil[i, 1])
    return gr


def segment_path(pts, ds=0, close=False):
    """Segment a path into equal length segments.

    Args:
        pts (array): N x 3 array of points
        ds (float, optional): length between points. Defaults to minimun in pts.
        close (bool, optional): Should the path close the points. Defaults to False.

    Returns:
        array: M x 3 array of points along the path
        array: M array of length along the path
    """
    # this will make random points have equal spaces along path
    # from scipy.interpolate import interp1d
    # assume path is x,y,z and the 0 axis is the segment
    # so pts[*,0] is x, pts[*,1] is y, pts[*,2] is z

    if close:
        pts = np.vstack([pts, pts[0]])
    dx = pts[1:] - pts[:-1]
    dsv = np.linalg.norm(dx, axis=1)
    if ds == 0:
        ds = dsv.min()
    s = np.insert(np.cumsum(dsv), 0, 0)
    slen = s[-1]  # length
    snew = np.linspace(0, slen, int(slen / ds))

    # use np.interp instead of scipy.interpolation.interp1d
    segs = np.column_stack([np.interp(snew, s, pts[:, i]) for i in range(3)])
    return segs, snew


@njit
def _loop_segmented_mutual(r, z, pts):
    # pts is array of n x 3 (x,y,z)
    # pts should contain first point at start AND end.
    # r, z is r & z of loop
    M = float(0)
    for i in range(pts.shape[0] - 1):
        midp = (pts[i, :] + pts[i + 1, :]) / 2
        delta = pts[i, :] - pts[i + 1, :]
        rs = math.sqrt(midp[0] ** 2 + midp[1] ** 2)
        zs = midp[2]
        rdphi = (delta[0] * midp[1] - delta[1] * midp[0]) / rs**2
        M += AGreen(r, z, rs, zs) * rdphi

    return M


@njit
def _segmented_segmented_mutual(pts1, pts2):
    Avecs = np.zeros_like(pts2)
    for i in range(pts1.shape[0] - 1):
        xyz = pts1[i]
        uvw = pts1[i + 1] - pts1[i]
        Avecs += ASegment(pts2, xyz, uvw)
    A_midps = (Avecs[1:] + Avecs[:-1]) / 2
    deltas = pts2[1:] - pts2[:-1]
    dots = (
        deltas[:, 0] * A_midps[:, 0]
        + deltas[:, 1] * A_midps[:, 1]
        + deltas[:, 2] * A_midps[:, 2]
    )
    return np.sum(dots)


@njit(parallel=True)
def mutual_filaments_segmented(fils, pts):
    """Mutual inductance between a set of axisymmetric filaments and a segmented path."""
    M = float(0)
    for i in prange(fils.shape[0]):
        M += fils[i, 2] * _loop_segmented_mutual(fils[i, 0], fils[i, 1], pts)
    return M


def M_filsol_path(fils, pts, nt, ds=0):
    """Mutual inductance between a set of axisymmetric filaments and a path from pts."""
    segs, _ = segment_path(pts, ds)
    return nt * mutual_filaments_segmented(fils, segs)


def M_path_path(pts1, pts2, ds1=0, ds2=0):
    """Mutual inductance between two pts paths."""
    segs1, _ = segment_path(pts1, ds1)
    segs2, _ = segment_path(pts2, ds2)
    return _segmented_segmented_mutual(segs1, segs2)


@njit(parallel=True)
def segmented_self_inductance(pts, s, a):
    """Self inductance of a segmented path by double integral.

    Args:
        pts (N x 3 array): points along the path
        s (array): length along the path
        a (float): radius of wire

    Returns:
        float: self inductance of the path

    Neumann's formula.. double curve integral of
    mu_0/(4*pi) * int * int ( dx1 . dx2 / norm(x1-x2)))
    do all points except where x1-x2 blows up.
    instead follow Dengler https://doi.org/10.7716/aem.v5i1.331
    which makes a approximation for that is good O(mu_0*a)
    lets just assume the thing is broken into small pieces and neglect end points

    this code doesn't work very well.. comparison test sorta fails
    phi = np.linspace(0,2*math.pi,100)
    test_xyz = np.array([[np.cos(p), np.sin(p), 0] for p in phi])
    L_seg = L_approx_path_rect(test_xyz, .01, .01, 1, .1)
    L_maxwell(1, .01, .01, 1), L_lyle6(1, .01, .01, 1), L_seg
    (6.898558527897293e-06, 6.8985981243869525e-06, 6.907313505254537e-06) # Y=1/4 hmmm...
    (6.898558527897293e-06, 6.8985981243869525e-06, 7.064366776069971e-06) # Y=1/2
    """
    ds = s[1] - s[0]  # the real ds and thus the real b
    dx = pts[1:] - pts[:-1]
    x = (pts[1:] + pts[:-1]) / 2
    npts = x.shape[0]
    slen = s[-1]
    b = ds / 2  # does not depend on ds, because of this correction.
    LS = 2 * slen * (np.log(2 * b / a) + 0.25)  # dengler equation 6 correction.
    # seems to work much better when its +.125
    L = 0  # numba REQUIRES prange parallel variable to start at 0 (bug!)
    for i in prange(npts):
        for j in range(npts):
            if i != j:
                L += np.dot(dx[i], dx[j]) / np.linalg.norm(x[i] - x[j])
    return 1e-7 * (L + LS)


def L_approx_path_rect(pts, w, h, n, ds=1):
    """Approximate self inductance of a path of points with a rectangular cross section.

    take a path of points n x 3, with a cross section b x c
    and approximate self inductance using Dengler
    """
    a = rectangle_GMD(w, h)  # get Maxwell mean radius to approximate "wire" radius
    ds *= a
    segs, s = segment_path(pts, ds)
    L = n**2 * segmented_self_inductance(segs, s, a)
    return L


# for some reason, these are slightly slower than the above.
# but they work for any shape r & z as long as they are the same
@guvectorize(
    ["void(float64[:,:], float64[:], float64[:], float64[:])"],
    "(p, q),(n),(n)->(n)",
    target="parallel",
)
def BrGreenFil(fil, r, z, gr):
    """Radial magnetic field greens functions from a set of filaments over a set of points.

    Args:
        fil (_type_): filament array (r,z,n)
        r (_type_): radial positions of points
        z (_type_): vertical positions of points
        gr (_type_): greens function values at points
    """
    for j in range(r.shape[0]):
        tmp = 0.0
        for i in range(len(fil)):
            tmp += fil[i, 2] * BrGreen(r[j], z[j], fil[i, 0], fil[i, 1])
        gr[j] = tmp


@guvectorize(
    ["void(float64[:,:], float64[:], float64[:], float64[:])"],
    "(p, q),(n),(n)->(n)",
    target="parallel",
)
def BzGreenFil(fil, r, z, gr):
    """Vertical magnetic field greens functions from a set of filaments over a set of points.

    Args:
        fil (_type_): filament array (r,z,n)
        r (_type_): radial positions of points
        z (_type_): vertical positions of points
        gr (_type_): greens function values at points
    """
    for j in range(r.shape[0]):
        tmp = 0.0
        for i in range(len(fil)):
            tmp += fil[i, 2] * BzGreen(r[j], z[j], fil[i, 0], fil[i, 1])
        gr[j] = tmp


@guvectorize(
    ["void(float64[:,:], float64[:], float64[:], float64[:])"],
    "(p, q),(n),(n)->(n)",
    target="parallel",
)
def AGreenFil(fil, r, z, gr):
    """Psi greens functions from a set of filaments over a set of points.

    Args:
        fil (_type_): filament array (r,z,n)
        r (_type_): radial positions of points
        z (_type_): vertical positions of points
        gr (_type_): greens function values at points
    """
    for j in range(r.shape[0]):
        tmp = 0.0
        for i in range(len(fil)):
            tmp += fil[i, 2] * AGreen(r[j], z[j], fil[i, 0], fil[i, 1])
        gr[j] = tmp


# cant jit this... meshgrid not allowed
def filament_coil(r, z, dr, dz, nt, nr, nz, theta=0):
    """Create an array of filaments, each with its own radius, height, and amperage.

    r : Major radius of coil center.
    z : Vertical center of coil.
    dr : Radial width of coil.
    dz : Height of coil.
    nt : number of turns in coil
    nr : Number of radial slices
    nz : Number of vertical slices
    theta : Rotation of coil about phi axis

    Returns:    Array of shape (nr*nz) x 3 of R, Z, N for each filament
    """
    sects = section_coil(r, z, dr, dz, nt, nr, nz, theta=theta)
    return sects[:, [0, 1, 4]]


@njit(parallel=True)
def sum_over_filaments(func, f1, f2):
    """Apply a function and sum over all combinations of two sets of filaments.

    Args:
        func (function): function to apply to each pair of filaments
        f1 (array): first filament array
        f2 (array): second filament array

    Returns:
        float: result of summing over all combinations of func(f1[i], f2[j])
    """
    result = float(0)
    for i in prange(f1.shape[0]):
        for j in range(f2.shape[0]):
            result += func(f1[i], f2[j])
    return result


@njit(parallel=True)
def mutual_inductance_of_filaments(f1, f2):
    """Mutual inductance of two sets of filaments.

    These should not be the same filament array,
    not setup to handle self inductance of filament.

    Args:
        f1 (array): first filament array
        f2 (array): second filament array

    Returns:
        float: Mutual inductance of filament sets in Henries
    """
    M = float(0)
    for i in prange(f1.shape[0]):
        for j in range(f2.shape[0]):
            M += mutual_inductance_fil(f1[i, :], f2[j, :])
    return M


@njit(parallel=True)
def vertical_force_of_filaments(f1, f2):
    """Vertical force between two sets of filaments.

    These should not be the same filament array,
    not setup to handle self inductance of filament.

    Args:
        f1 (array): first filament array
        f2 (array): second filament array

    Returns:
        float: Vertical force in Newtons/Amp^2
    """
    F = float(0)
    for i in prange(f1.shape[0]):
        for j in range(f2.shape[0]):
            F += vertical_force_fil(f1[i, :], f2[j, :])
    return F


@njit(parallel=True)
def radial_force_of_filaments(f1, f2):
    """Radial force between two sets of filaments.

    These should not be the same filament array,
    not setup to handle self inductance of filament.

    Args:
        f1 (array): first filament array
        f2 (array): second filament array

    Returns:
        float: Radial force in Newtons/Amp^2
    """
    F = float(0)
    for i in prange(f1.shape[0]):
        for j in range(f2.shape[0]):
            F += radial_force_fil(f1[i, :], f2[j, :])
    return F



"""Test filamentation and inductance of LDX coils."""

import unittest

import coverage_env  # noqa: F401

from inductance.coils import Coil, CompositeCoil


class TestLDXInductance(unittest.TestCase):
    """Test the inductance of the LDX coils.

    We know the coil weighed about 590kg, and had about 0.4H inductance.
    """

    def setUp(self) -> None:
        """Set up the LDX coils."""
        (
            self.HTS_LC,
            self.LC,
            self.CC,
            self.F1,
            self.F2,
            self.F3,
            self.FC,
        ) = define_ldx_coils()
        return super().setUp()

    def test_self_inductances(self):
        """Test the self-inductances routines."""
        self.assertAlmostEqual(self.LC.L_Maxwell(), 0.005397519485316731)
        self.assertAlmostEqual(self.LC.L_Lyle4(), 0.005623887363536865)
        self.assertAlmostEqual(self.LC.L_Lyle6(), 0.005626208118367906)
        self.assertAlmostEqual(self.LC.L_Lyle6A(), 0.005626208118367906)
        self.assertAlmostEqual(self.LC.L_filament(), 0.005625117051066614)

        self.assertAlmostEqual(self.CC.L_Maxwell(), 85.91637858501646)
        self.assertAlmostEqual(self.CC.L_Lyle4(), 90.90254315310752)
        self.assertAlmostEqual(self.CC.L_Lyle6(), 90.90927053789774)
        self.assertAlmostEqual(self.CC.L_Lyle6A(), 90.90927053789773)
        self.assertAlmostEqual(self.CC.L_filament(), 90.89118772211005)

    def test_mutual_inductances(self):
        """Test the mutual inductances routines."""
        MFL = self.FC.M_filament(self.LC)
        MFC = self.FC.M_filament(self.CC)
        self.assertAlmostEqual(MFL, 0.611729e-3)
        self.assertAlmostEqual(MFC, 1.686291576361124)

    def test_fcoil_selfinductance(self):
        """Test the self-inductance of the F coils."""
        LF1 = self.F1.L_Lyle6()
        LF2 = self.F2.L_Lyle6()
        LF3 = self.F3.L_Lyle6()
        MF12 = self.F1.M_filament(self.F2)
        MF13 = self.F1.M_filament(self.F3)
        MF23 = self.F2.M_filament(self.F3)
        LFC = LF1 + LF2 + LF3 + 2 * MF12 + 2 * MF13 + 2 * MF23
        self.assertAlmostEqual(LFC, 0.38692937382836284)

    def test_levitation_force(self):
        """Test the levitation force."""
        FZ = self.FC.Fz_filament(self.LC) / 9.81
        self.assertAlmostEqual(FZ, 589.263842397435)


def define_ldx_coils():
    """Define the LDX coils."""
    HTS_LC = {}
    HTS_LC["r1"] = 0.41 / 2
    HTS_LC["r2"] = 1.32 / 2
    HTS_LC["z1"] = 1.610 - 0.018 / 2
    HTS_LC["z2"] = 1.610 + 0.018 / 2
    HTS_LC["nt"] = 2796
    HTS_LC["at"] = 2796 * 105
    HTS_LC = Coil.from_dict(HTS_LC)
    HTS_LC.filamentize(30, 2)

    LC = {}
    LC["r1"] = 0.246
    LC["r2"] = 0.70
    LC["z1"] = 1.525
    LC["z2"] = LC["z1"] + 0.1
    LC["nt"] = 80
    LC["at"] = 3500 * 80
    LC = Coil.from_dict(LC)
    LC.filamentize(20, 4)

    CC = {}
    CC["r1"] = 0.645
    CC["r2"] = 0.787
    CC["z1"] = -0.002 - 0.750 / 2
    CC["z2"] = -0.002 + 0.750 / 2
    CC["nt"] = 8388
    CC["at"] = 8388 * 420
    CC = Coil.from_dict(CC)
    CC.filamentize(3, 7)

    F1 = {}
    F1["r1"] = 0.2717 - 0.01152 / 2
    F1["r2"] = 0.2717 + 0.01152 / 2
    F1["z1"] = -0.0694 / 2
    F1["z2"] = +0.0694 / 2
    F1["nt"] = 26.6
    F1["at"] = 26.6 * 1629
    F1 = Coil.from_dict(F1)
    F1.filamentize(2, 4)

    F2 = {}
    F2["r1"] = 0.28504 - 0.01508 / 2
    F2["r2"] = 0.28504 + 0.01508 / 2
    F2["z1"] = -0.125 / 2
    F2["z2"] = +0.125 / 2
    F2["nt"] = 81.7
    F2["at"] = 81.7 * 1629
    F2 = Coil.from_dict(F2)
    F2.filamentize(3, 7)

    F3 = {}
    F3["r1"] = 0.33734 - 0.08936 / 2
    F3["r2"] = 0.33734 + 0.08936 / 2
    F3["z1"] = -0.1615 / 2
    F3["z2"] = +0.1615 / 2
    F3["nt"] = 607.7
    F3["at"] = 607.7 * 1629
    F3 = Coil.from_dict(F3)
    F3.filamentize(10, 15)

    FC = CompositeCoil([F1, F2, F3])

    return HTS_LC, LC, CC, F1, F2, F3, FC


if __name__ == "__main__":
    unittest.main()



"""Self nductance calculations for coils.

author: Darren Garnier <garnier@mit.edu>

basic equations come from analytic approximations of old.
tested in real life with the LDX Fcoil / Ccoil / Lcoil system.

One from Maxwell himself, and better ones from:

Lyle, T. R.  "On the Self-inductance of Circular Coils of
    Rectangular Section".  Roy. Soc. London. A.  V213 (1914) pp 421-435.
    https://doi.org/10.1098/rsta.1914.0009

Unfortunately, Lyle doesn't work that well with large dz/R coils.  Other
approximations are also included.

This code now uses numba to do just-in-time compiliation and parallel execution
to greatly increase speed. Also requires numba-scipy for elliptical functions.
numba-scipy can be fragile and sometimes needs to be "updated" before installation
to the newest version of numba.
"""

import math

import numpy as np

from ._numba import njit, prange
from .elliptics import ellipke
from .filaments import mutual_inductance_fil
from .mutual import lyle_equivalent_subcoil_filaments, mutual_lyles_method
from .utils import _lyle_terms, section_coil

MU0 = 4e-7 * math.pi  # permeability of free space


@njit
def L_maxwell(r, dr, dz, n):
    """Self inductance of a rectangular coil with constant current density by Maxwell.

    Args:
        r (float): coil centerline radius
        dr (float): coil radial width
        dz (float): coil height
        n (int): number of turns

    Returns:
        float: coil self inductance in Henrys
    """
    a = float(r)
    b = float(dz)
    c = float(dr)
    d, u, v, w, wp, phi, GMD = _lyle_terms(b, c)
    L = MU0 * (n**2) * a * (math.log(8 * a / GMD) - 2)
    return L


@njit
def L_round(r, a, n):
    """Self inductance of a round conductor coil with constant current density.

    Args:
        r (float): coil centerline radius
        a (float): coil conductor radius
        n (int): number of turns

    Returns:
        float: coil self inductance in Henrys
    """
    L = MU0 * (n**2) * r * (math.log(8 * r / a) - 1.75)
    return L


@njit
def L_hollow_round(r, a, n):
    """Self inductance of a round conductor coil with skin current.

    Args:
        r (float): coil centerline radius
        a (float): coil conductor radius
        n (int): number of turns

    Returns:
        float: coil self inductance in Henrys
    """
    L = MU0 * (n**2) * r * (math.log(8 * r / a) - 2)
    return L


@njit
def L_lyle4(r, dr, dz, n):
    """Self inductance of a rectangular coil via Lyle to 4th order, Eq3.

    Args:
        r (float): coil centerline radius
        dr (float): coil radial width
        dz (float): coil height
        n (int): number of turns

    Returns:
        float: coil self inductance in Henrys
    """
    a = float(r)
    b = float(dz)
    c = float(dr)
    d, u, v, w, wp, phi, GMD = _lyle_terms(b, c)
    p2 = 1 / (2**5 * 3 * d**2) * (3 * b**2 + c**2)
    q2 = (
        1
        / (2**5 * 3 * d**2)
        * (
            1 / 2 * b**2 * u
            - 1 / 10 * c**2 * v
            - 16 / 5 * b**2 * w
            - (3 * b**2 + c**2) * phi
            + 69 / 20 * b**2
            + 221 / 60 * c**2
        )
    )
    p4 = 1 / (2**11 * 3**2 * 5 * d**4) * (-90 * b**4 + 105 * (b * c) ** 2 + 22 * c**4)
    q4 = (
        1
        / (2**11 * 3**2 * 5 * d**4)
        * (
            -(-90 * b**4 + 105 * (b * c) ** 2 + 22 * c**4) * phi
            - 69 / 28 * c**4 * v
            - u / 4 * (115 * b**4 - 480 * (b * c) ** 2)
            + 2**8 * w / 7 * (6 * b**4 - 7 * (b * c) ** 2)
            - 1 / (2**3 * 5 * 7) * (36590 * b**4 - 2035 * (b * c) ** 2 - 11442 * c**4)
        )
    )

    ML = np.log(8 * a / GMD)

    # equation #3

    eq3 = (
        MU0
        * (n**2)
        * a
        * (ML - 2 + (d / a) ** 2 * (p2 * ML + q2) + (d / a) ** 4 * (p4 * ML + q4))
    )
    return eq3


# equation 4.. slightly different result... not sure which is better.
# equation 3 above matches what I did for the 6th order.
# and it also seems to match the 4th order in the paper.
# and in other papers with examples


@njit
def L_lyle4_eq4(r, dr, dz, n):
    """Self inductance of a rectangular coil via Lyle to 4th order, Eq4.

    this doesn't give quite the same answer as eq3 above.
    and it doesn't seem to work as well.

    Args:
        r (float): coil centerline radius
        dr (float): coil radial width
        dz (float): coil height
        n (int): number of turns

    Returns:
        float: coil self inductance in Henrys
    """
    a = float(r)
    b = float(dz)
    c = float(dr)
    d, u, v, w, wp, phi, GMD = _lyle_terms(b, c)
    p2 = 1 / (2**5 * 3 * d**2) * (3 * b**2 + c**2)
    q2 = (
        1
        / (2**5 * 3 * d**2)
        * (
            1 / 2 * b**2 * u
            - 1 / 10 * c**2 * v
            - 16 / 5 * b**2 * w
            - (3 * b**2 + c**2) * phi
            + 69 / 20 * b**2
            + 221 / 60 * c**2
        )
    )
    p4 = 1 / (2**11 * 3**2 * 5 * d**4) * (-90 * b**4 + 105 * (b * c) ** 2 + 22 * c**4)
    q4 = (
        1
        / (2**11 * 3**2 * 5 * d**4)
        * (
            -(-90 * b**4 + 105 * (b * c) ** 2 + 22 * c**4) * phi
            - 69 / 28 * c**4 * v
            - u / 4 * (115 * b**4 - 480 * (b * c) ** 2)
            + 2**8 * w / 7 * (6 * b**4 - 7 * (b * c) ** 2)
            - 1 / (2**3 * 5 * 7) * (36590 * b**4 - 2035 * (b * c) ** 2 - 11442 * c**4)
        )
    )

    m1 = p2
    n1 = -(p2 + q2)
    m2 = p4
    n2 = -(p4 + q4) + 1 / 2 * (m1 - n1) ** 2
    n3 = (m1 - n1) * (m2 - n2 - 1 / 6 * (m1 - n1) * (m1 + 2 * n1))

    A = a * (1 + m1 * (d / a) ** 2 + m2 * (d / a) ** 4)
    R = GMD * (1 + n1 * (d / a) ** 2 + n2 * (d / a) ** 4 + n3 * (d / a) ** 6)

    eq4 = MU0 * (n**2) * A * (np.log(8 * A / R) - 2)
    return eq4


@njit
def L_lyle6(r, dr, dz, n):
    """Self inductance of a rectangular coil via Lyle to 6th order.

    Args:
        r (float): coil centerline radius
        dr (float): coil radial width
        dz (float): coil height
        n (int): number of turns

    Returns:
        float: coil self inductance in Henrys
    """
    a = float(r)
    b = float(dz)
    c = float(dr)
    d, u, v, w, ww, phi, GMD = _lyle_terms(b, c)
    bd2 = (b / d) ** 2
    cd2 = (c / d) ** 2
    da2 = (d / a) ** 2
    ML = np.log(8 * a / d)

    # after further reduction in mathematica... all the terms.
    f = (
        ML
        + (1 + u + v - 8 * (w + ww)) / 12.0  # 0th order in d/a
        + (
            da2
            * (
                cd2 * (221 + 60 * ML - 6 * v)
                + 3 * bd2 * (69 + 60 * ML + 10 * u - 64 * w)
            )
        )
        / 5760.0  # 2nd order
        + (
            da2**2
            * (
                2 * cd2**2 * (5721 + 3080 * ML - 345 * v)
                + 5 * bd2 * cd2 * (407 + 5880 * ML + 6720 * u - 14336 * w)
                - 10 * bd2**2 * (3659 + 2520 * ML + 805 * u - 6144 * w)
            )
        )
        / 2.58048e7  # 4th order
        + (
            da2**3
            * (
                3 * cd2**3 * (4308631 + 86520 * ML - 10052 * v)
                - 14 * bd2**2 * cd2 * (617423 + 289800 * ML + 579600 * u - 1474560 * w)
                + 21 * bd2**3 * (308779 + 63000 * ML + 43596 * u - 409600 * w)
                + 42 * bd2 * cd2**2 * (-8329 + 46200 * ML + 134400 * u - 172032 * w)
            )
        )
        / 1.73408256e10  # 6th order
    )
    L = MU0 * (n**2) * a * f
    # print("Lyle6 r: %.4g, dr: %.4g, dz: %4g, n: %d, L: %.8g"%(a,c,b,n,L))
    return L


@njit
def dLdR_lyle6(r, dr, dz, n):
    """Radial derivative of self inductance of a rectangular coil via Lyle to 6th order.

    Args:
        r (float): coil centerline radius
        dr (float): coil radial width
        dz (float): coil height
        n (int): number of turns

    Returns:
        float: radial derivative of inductance in Henrys/meter
    """
    a = float(r)
    b = float(dz)
    c = float(dr)
    d, u, v, w, ww, phi, GMD = _lyle_terms(b, c)
    bd2 = (b / d) ** 2
    cd2 = (c / d) ** 2
    da2 = (d / a) ** 2

    ML = np.log(8 * a / d)

    f = (
        ML
        + (13 + u + v - 8 * w - 8 * ww) / 12.0  # zero
        + (
            da2
            * (
                cd2 * (-161 - 60 * ML + 6 * v)
                - 3 * bd2 * (9 + 60 * ML + 10 * u - 64 * w)
            )
        )
        / 5760.0  # 2nd order
        + (
            da2**2
            * (
                -2 * cd2**2 * (14083 + 9240 * ML - 1035 * v)
                - 15 * bd2 * cd2 * (-1553 + 5880 * ML + 6720 * u - 14336 * w)
                + 30 * bd2**2 * (2819 + 2520 * ML + 805 * u - 6144 * w)
            )
        )
        / 2.58048e7  # 4th order
        + (
            da2**3
            * (
                -3 * cd2**3 * (4291327 + 86520 * ML - 10052 * v)
                + 14 * bd2**2 * cd2 * (559463 + 289800 * ML + 579600 * u - 1474560 * w)
                - 21 * bd2**3 * (296179 + 63000 * ML + 43596 * u - 409600 * w)
                - 42 * bd2 * cd2**2 * (-17569 + 46200 * ML + 134400 * u - 172032 * w)
            )
        )
        / 3.46816512e9  # 6th order
    )

    dLdR = 4e-7 * np.pi * (n**2) * f
    return dLdR


@njit
def L_lyle6_appendix(r, dr, dz, n):
    """Self inductance of a rectangular coil via Lyle to 6th order, appendix.

    Args:
        r (float): coil centerline radius
        dr (float): coil radial width
        dz (float): coil height
        n (int): number of turns

    Returns:
        float: coil self inductance in Henrys
    """
    a = float(r)
    b = float(dz)
    c = float(dr)
    d, u, v, w, wp, phi, GMD = _lyle_terms(b, c)

    p6 = (
        1
        / (2**16 * 3 * 5 * 7 * d**6)
        * (525 * b**6 - 1610 * b**4 * c**2 + 770 * b**2 * c**4 + 103 * c**6)
    )
    q6 = (
        1
        / (2**16 * 3 * 5 * 7 * d**6)
        * (
            0
            + (3633 / 10 * b**6 - 3220 * b**4 * c**2 + 2240 * b**2 * c**4) * u
            - (359 / 30) * c**6 * v
            - 2**11 * (5 / 3 * b**6 - 4 * b**4 * c**2 + 7 / 5 * b**2 * c**4) * w
            + 2161453 / (2**3 * 3 * 5 * 7) * b**6
            - 617423 / (2**2 * 3**2 * 5) * b**4 * c**2
            - 8329 / (2**2 * 3 * 5) * b**2 * c**4
            + 4308631 / (2**3 * 3 * 5 * 7) * c**6
        )
    )

    # just add the correction to the 4th order solution
    L6 = L_lyle4(r, dr, dz, n) + 4e-7 * np.pi * (n**2) * a * (d / a) ** 6 * (
        p6 * np.log(8 * a / d) + q6
    )
    # print("Lyle6A r: %.4g, dr: %.4g, dz: %4g, n: %d, L: %.8g"%(a,c,b,n,L6))
    return L6


@njit
def L_lyle_sectioning(r, dr, dz, nt, nr, nz):
    """Self inductance by sectioning, Lyle's method, and Lyle's 4th order.

    Args:
        r (float): coil centerline radius
        dr (float): coil radial width
        dz (float): coil height
        nt (float): number of turns
        nr (int): number of radial sections
        nz (int): number of axial sections

    Returns:
        float: coil self inductance in Henrys
    """
    L = 0
    subcoils = section_coil(r, dr, dz, nt, nt, nr, nz)
    fils = lyle_equivalent_subcoil_filaments(subcoils)

    for i in range(nr * nz):
        for j in range(nr * nz):
            if i != j:
                L += mutual_lyles_method(*fils[i], *fils[j])
        L += 2 * L_lyle6(*subcoils[i])
    return L


@njit
def L_long_solenoid_butterworth(r, dr, dz, n):
    """Self inductance of a long solenoid by Butterworth's formula.

    As written in Grover, Bulletin of the Bureau of Standards, Vol. 14 pg. 558
    https://nvlpubs.nist.gov/nistpubs/bulletin/14/nbsbulletinv14n4p537_A2b.pdf

    Original S Butterworth 1914 Proc. Phys. Soc. London 27 371

    Applies when dz > 2*r.

    Args:
        r (float): coil centerline radius
        dr (float): coil radial width
        dz (float): coil height
        n (int): number of turns

    Returns:
        float: coil self inductance in Henrys
    """
    a = float(r)
    b = float(dz)
    c = float(dr)
    # L = 4e-7 * np.pi * (n**2) * a / b

    k2 = 4 * a**2 / (4 * a**2 + b**2)
    kp2 = b**2 / (4 * a**2 + b**2)
    k = np.sqrt(k2)
    kp = np.sqrt(kp2)

    # assume dz > 2*r
    ell = k2 / ((1 + kp) * (1 + np.sqrt(kp)) ** 2)
    q = ell / 2 + 2 * (ell / 2) ** 5 + 15 * (ell / 2) ** 9  # + ....

    delta = 2 * q - 2 * q**4 + 2 * q**9
    gamma = q - 4 * q**4 + 9 * q**9
    beta = q**2 + 3 * q**6 + 6 * q**12
    alpha = q**2 + q**6 + q**12

    K = (  # K
        2.0
        / (3 * (1 - delta) ** 2)
        * (1 + 8 * beta / (1 + alpha) + kp2 / k2 * 8 * gamma / (1 - delta))
        - 4 / (3 * np.pi) * k / kp
    )

    DL_L1 = (  # delta L/L_1, eq 29A
        -1
        / 3
        * (c / a)
        * (
            1
            - c / (4 * a)
            - 1 / (2 * np.pi) * (c / b) * (np.log(8 * a / c) - 23 / 12)
            + 1
            / (160 * np.pi)
            * (c / a) ** 3
            * (a / b)
            * (np.log(8 * a / c) - 1.0 / 20)
            - 1
            / 4
            * (c * a / b**2)
            * (1 - 7 / 4 * (a / b) ** 2 + 17 / 4 * (a / b) ** 4)
            - 1 / 96 * (c / a) ** 3 * (a / b) ** 2 * (1 - 39 / 10 * (a / b) ** 2)
        )
    )
    L1 = L_lorentz(r, dr, dz, n)

    return L1 * (K + DL_L1)


@njit
def _L_thin_wall_babic_akyel(r, _dr, dz, n):
    """Self inductance thin wall solenoid by Babic and Akyel's formula.

    Follow formulae from:
    S. Babic and C. Akyel, "Improvement in calculation of the self- and mutual inductance
    of thin-wall solenoids and disk coils," in IEEE Transactions on Magnetics,
    vol. 36, no. 4, pp. 1970-1975, July 2000, doi: 10.1109/TMAG.2000.875240.

    this _is_ be the same as Lorentz formula..

    Args:
        r (float): coil centerline radius
        _dr (float): coil radial width (ignored)
        dz (float): coil height
        n (int): number of turns

    Returns:
        float: coil self inductance in Henrys
    """
    a = float(dz) / 2
    # R1 = float(r) - float(dr) / 2
    # R2 = float(r) + float(dr) / 2
    # alpha = R2 / R1
    # lets double check why I'm not using alpha...
    beta = a / r

    k2 = 1 / (1 + beta**2)
    elk, ele = ellipke(k2)
    k = np.sqrt(k2)
    tk1 = 4.0 / (3 * np.pi * beta * k**3) * ((2 * k2 - 1) * ele + (1 - k2) * elk - k**3)
    eq8 = MU0 * (n**2) * r * np.pi / (2 * beta) * tk1  # eq 8
    return eq8


@njit
def L_lorentz(r, _dr, dz, n):
    """Self inductance of a thin wall solenoid by Lorentz's formula.

    Given in:
    Rosa and Grover, Formulas and Tables for the Calculation of Mutual and Self-Inductance
    https://nvlpubs.nist.gov/nistpubs/bulletin/08/nbsbulletinv8n1p1_A2b.pdf
    or
    https://www.jstor.org/stable/pdf/24521000.pdf
    (Formula 72 on page 118)

    Args:
        r (float): coil centerline radius
        _dr (float): coil radial width (ignored)
        dz (float): coil height
        n (int): number of turns

    Returns:
        float: coil self inductance in Henrys
    """
    beta = dz / (2 * r)
    k2 = 1 / (1 + beta**2)
    k3 = k2**1.5
    beta2 = beta**2
    elK, elE = ellipke(k2)

    f = 2 / 3 / beta2 * (((2 * k2 - 1) * elE + (1 - k2) * elK) / k3 - 1)
    Ls = MU0 * (n**2) * r * f
    return Ls


@njit(parallel=True)
def self_inductance_by_filaments(f, conductor="round", a=0.01, dr=0.01, dz=0.01):
    """Self inductance of filament set.

    Args:
        f (array): first filament array
        conductor (str, optional): conductor shape. Defaults to "round".
        a (float, optional): conductor radius. Defaults to 0.01.
        dr (float, optional): conductor radial width. Defaults to 0.01
        dz (float, optional): conductor vertical height. Defaults to 0.01

    Returns:
        float: self inductance of filament set in Henries
    """
    L = float(0)
    for i in prange(f.shape[0]):
        for j in range(f.shape[0]):
            if i != j:
                L += mutual_inductance_fil(f[i, :], f[j, :])
        if conductor == "round":
            L += L_round(f[i, 0], a, f[i, 2])
        elif conductor == "hollow_round":
            L += L_hollow_round(f[i, 0], a, f[i, 2])
        elif conductor == "rect":
            L += L_lyle6(f[i, 0], dr, dz, f[i, 2])
    return L



"""Utility functions for inductance calculations.

This module contains utility functions for inductance calculations.
"""

import math

import numpy as np

from ._numba import njit


@njit
def _lyle_terms(b, c):
    """Helper function for Lyle's self-inductance equations.

    Args:
        b (float): length of coil
        c (float): radial width of coil

    Returns:
        : _description_
    """
    # the basic formulae and gives inaccurate results when b/c ~ 1e6 or 1e-6
    # phi should approach 1.5 and GMD should approach (b+c)*exp(-1.5)
    # so special case the ends, and for small and large b/c
    # will try to use limit functions and series expansions
    boc2 = (b / c) ** 2
    if b == 0:
        u, v, w, p = 0, 1, 0, 1
    elif c == 0:
        u, v, w, p = 1, 0, 1, 0
    elif boc2 < 1e-8:
        u = -boc2 * math.log(boc2) + boc2**2 - boc2**3 / 2
        v = 1 - boc2 / 2 + boc2**2 / 3
        w = math.pi / 2 * (b / c) - boc2 + boc2**2 / 3
        p = 1 - boc2 / 3 + boc2**2 / 5
    elif boc2 > 1e8:
        cob2 = (c / b) ** 2
        u = 1 - cob2 / 2 + cob2**2 / 3
        v = -cob2 * math.log(cob2) + cob2**2 - cob2**3 / 2
        w = 1 - cob2 / 3 + cob2**2 / 5  # taylor series
        p = math.pi / 2 * (c / b) - cob2 + cob2**2 / 3  # laurent series
    else:
        u = ((b / c) ** 2) * math.log(1 + (c / b) ** 2)
        v = ((c / b) ** 2) * math.log(1 + (b / c) ** 2)
        w = (b / c) * math.atan2(c, b)
        p = (c / b) * math.atan2(b, c)

    d = np.sqrt(b**2 + c**2)  # diagnonal length
    phi = (u + v + 25) / 12 - 2 * (w + p) / 3
    GMD = d * np.exp(-phi)  # geometric mean radius of section GMD

    return d, u, v, w, p, phi, GMD


def rectangle_GMD(dr, dz):
    """Geometric mean radius of a rectangle.

    Args:
        dr (float): width of rectangle
        dz (float): height of rectangle

    Returns:
        float: GMD of rectangle
    """
    return _lyle_terms(dz, dr)[-1]


def section_coil(r, z, dr, dz, nt, nr, nz, theta=0):
    """Create an array of subcoils.

    Each coil will have its with its own radius, height,
    number of turns.

    Args:
        r  (float): Major radius of coil center.
        z  (float): Vertical center of coil.
        dr (float): Radial width of coil.
        dz (float): Height of coil.
        nt (float): number of turns in coil
        nr (float): Number of radial slices
        nz (float): Number of vertical slices
        theta (float): Rotation angle in radians

    Returns:
        (np.ndarray) : Array of shape (nr*nz) x 5 of r, z, dr, dz, nt)
        for each section
    """
    rd = np.linspace(-dr * (nr - 1) / nr / 2, dr * (nr - 1) / nr / 2, nr)
    zd = np.linspace(-dz * (nz - 1) / nz / 2, dz * (nz - 1) / nz / 2, nz)

    Rg, Zg = np.meshgrid(rd, zd)

    R = r + Rg * np.cos(theta) - Zg * np.sin(theta)
    Z = z + Rg * np.sin(theta) + Zg * np.cos(theta)

    DR = np.full_like(R, dr / nr)
    DZ = np.full_like(R, dz / nz)
    NT = np.full_like(R, float(nt) / (nr * nz))
    TH = np.full_like(R, theta)

    return np.dstack([R, Z, DR, DZ, NT, TH]).reshape(nr * nz, 6)



"""Numba stubs.
Copied from pandas.
"""

# pyright: reportIncompleteStub = false
from typing import TYPE_CHECKING, Any, Callable, Literal, TypeVar, overload

if TYPE_CHECKING:
    import numba

FuncType = Callable[..., Any]
F = TypeVar("F", bound=FuncType)


def __getattr__(name: str) -> Any: ...  # incomplete
@overload
def jit(signature_or_function: F) -> F: ...
@overload
def jit(
    signature_or_function: str
    | list[str]
    | numba.core.types.abstract.Type
    | list[numba.core.types.abstract.Type] = ...,
    locals: dict = ...,  # TODO: Mapping of local variable names to Numba types
    cache: bool = ...,
    pipeline_class: numba.compiler.CompilerBase = ...,
    boundscheck: bool | None = ...,
    *,
    nopython: bool = ...,
    forceobj: bool = ...,
    looplift: bool = ...,
    error_model: Literal["python", "numpy"] = ...,
    inline: Literal["never", "always"] | Callable = ...,
    # TODO: If a callable is provided it will be called with the call expression
    # node that is requesting inlining, the caller's IR and callee's IR as
    # arguments, it is expected to return Truthy as to whether to inline.
    target: Literal["cpu", "gpu", "npyufunc", "cuda"] = ...,  # deprecated
    nogil: bool = ...,
    parallel: bool = ...,
) -> Callable[[F], F]: ...

njit = jit
generated_jit = jit



"""Try to import numba."""

import os

# try to enable use without numba.  Those with guvectorize will not work.

try:
    from numba import guvectorize, jit, njit, prange

    # if numba is diabled, redfine the jit decorator to do nothing
    if os.getenv("NUMBA_DISABLE_JIT", "0") == "1":  # pragma: no cover
        raise ImportError

except ImportError:  # pragma: no cover
    from inspect import currentframe, getframeinfo
    from warnings import warn, warn_explicit

    if not os.getenv("COVERAGE_RUN", ""):  # pragma: no cover
        _WARNING = "Numba acceleration disabled. Some API will not be available."
        warn(_WARNING, RuntimeWarning)

    def _jit(*args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            # called as @decorator
            return args[0]
        else:
            # called as @decorator(*args, **kwargs)
            return lambda f: f

    def _guvectorize(*args, **kwds):
        def fake_decorator(f):
            if not os.getenv("COVERAGE_RUN", ""):  # pragma: no cover
                warning = f"{f.__name__} requires Numba JIT."
                finfo = getframeinfo(currentframe().f_back)  # type: ignore
                warn_explicit(warning, RuntimeWarning, finfo.filename, finfo.lineno)
            return lambda f: None

        return fake_decorator

    guvectorize = _guvectorize
    njit = _jit
    prange = range
    jit = _jit


__all__ = ["guvectorize", "jit", "njit", "prange"]
